# Copyright (c) 2012-2016 Hewlett Packard Enterprise Development Company LP
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os

import six

from requestbuilder import Arg


class RegionConfigurableMixin(object):
    """
    A mixin that allows the user to specify which user/region names to use
    for the configuration via a --region arg and, if a 'REGION_ENVVAR' class
    variable is set, an environment variable as well.  The included
    update_config_view method actually reads these data and updates
    self.config to point to them.
    """

    ARGS = [Arg('--region', metavar='USER@REGION', route_to=None,
                help=('region and/or user names to search when looking up '
                      'config file data'))]

    def update_config_view(self, region=None, user=None):
        # If the caller specified something directly we overwrite
        # whatever may be there unconditionally.
        if region:
            self.config.region = region
        if user:
            self.config.user = user

        # Otherwise, different sources of user/region info can override
        # only one of the pair, so we only overwite them individually,
        # and then only when no values are already set.

        # self.args gets highest precedence
        if self.args.get('region'):
            self.__setdefault_view_attrs(
                '--region CLI option "{0}"'.format(self.args['region']),
                *self.__parse_region(self.args['region']))
        # Environment comes next
        region_envvar = getattr(self, 'REGION_ENVVAR', None)
        if isinstance(region_envvar, (list, tuple)):
            for var in region_envvar:
                if os.getenv(var):
                    self.__setdefault_view_attrs(
                        'environment variable "{0}"'.format(var),
                        *self.__parse_region(os.getenv(var)))
                    break
        elif isinstance(region_envvar, six.string_types):
            if os.getenv(region_envvar):
                self.__setdefault_view_attrs(
                    'environment variable "{0}"'.format(region_envvar),
                    *self.__parse_region(os.getenv(region_envvar)))
        # Default region from the config file
        self.__setdefault_view_attrs(
            '"default-region" global config option',
            region=self.config.get_global_option('default-region'))
        # We've gone through all possible means of region selection, so
        # if no user is yet selected look it up in the region's config.
        user, section = self.config.get_region_option2('user')
        self.__setdefault_view_attrs(
            '"user" config option for region "{0}"'.format(section),
            user=user)

    def __setdefault_view_attrs(self, cause, user=None, region=None):
        if region and not self.config.region:
            self.log.info('selected region "%s" from %s', region, cause)
            self.config.region = region
        if user and not self.config.user:
            self.log.info('selected user "%s" from %s', user, cause)
            self.config.user = user

    @staticmethod
    def __parse_region(regionish):
        """
        Given a string with pattern "[USER@][REGION]", return the user
        and region names that that string represents, if any, and None
        for the values it does not represent.

        Examples:
         - ""          -> (None, None)
         - "spam"      -> (None, "spam")
         - "eggs@"     -> ("eggs", None)
         - "eggs@spam" -> ("eggs", "spam")
        """

        if not regionish:
            return None, None
        if regionish.endswith('@'):
            return regionish.rstrip('@'), None
        elif '@' in regionish:
            return regionish.split('@', 1)
        else:
            return None, regionish


# Compatibility with requestbuilder < 0.3
from .formatting import TabifyingMixin
from .progress import FileTransferProgressBarMixin



# Copyright (c) 2012-2016 Hewlett Packard Enterprise Development LP
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import absolute_import

import logging
import warnings


class BaseAuth(object):
    '''
    Basis for all authentication

    This class does nothing on its own.  It is up to you to implement the
    necessary functions to effect an authentication scheme.
    '''
    ARGS = []

    def __init__(self, config, loglevel=None, **kwargs):
        self.args = kwargs
        self.config = config
        self.log = logging.getLogger(self.__class__.__name__)
        if loglevel is not None:
            self.log.level = loglevel

    def configure(self):
        # TODO:  rename this to setup
        pass

    def apply_to_request(self, request, service):
        pass

    def bind_to_service(self, service):
        def wrapped_apply_to_request(req):
            return self.apply_to_request(req, service) or req
        return wrapped_apply_to_request


# Compatibility with requestbuilder < 0.3
from .aws import HmacKeyAuth
from .aws import HmacV1Auth as S3RestAuth
from .aws import QueryHmacV2Auth as QuerySigV2Auth



# Copyright (c) 2012-2015, Eucalyptus Systems, Inc.
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import absolute_import

import base64
import calendar
import datetime
import email.utils
import hashlib
import hmac
import os
import re
import tempfile
import time
import warnings

import six
import six.moves.urllib_parse as urlparse

from requestbuilder import Arg
from requestbuilder.auth import BaseAuth
from requestbuilder.exceptions import AuthError


ISO8601 = '%Y-%m-%dT%H:%M:%SZ'
ISO8601_BASIC = '%Y%m%dT%H%M%SZ'


class HmacKeyAuth(BaseAuth):
    '''
    Basis for AWS HMAC-based authentication
    '''
    ARGS = [Arg('-I', '--access-key-id', dest='key_id', metavar='KEY_ID'),
            Arg('-S', '--secret-key', dest='secret_key', metavar='KEY'),
            Arg('--security-token', dest='security_token', metavar='TOKEN')]

    @classmethod
    def from_other(cls, other, **kwargs):
        kwargs.setdefault('loglevel', other.log.level)
        kwargs.setdefault('key_id', other.args.get('key_id'))
        kwargs.setdefault('secret_key', other.args.get('secret_key'))
        kwargs.setdefault('security_token', other.args.get('security_token'))
        kwargs.setdefault('credential_expiration',
                          other.args.get('credential_expiration'))
        new = cls(other.config, **kwargs)
        new.configure()
        return new

    def configure(self):
        self.__populate_auth_args()

        if not self.args.get('key_id'):
            raise AuthError('missing access key ID; please supply one with -I')
        if not self.args.get('secret_key'):
            raise AuthError('missing secret key; please supply one with -S')
        if self.args.get('credential_expiration'):
            expiration = None
            for fmt in ('%Y-%m-%dT%H:%M:%S.%fZ', '%Y-%m-%dT%H:%M:%SZ'):
                try:
                    expiration = datetime.datetime.strptime(
                        self.args['credential_expiration'], fmt)
                    break
                except ValueError:
                    continue
            else:
                self.log.warn(
                    'failed to parse credential expiration time '
                    '\'{0}\'; proceeding without validation'
                    .format(self.args['credential_expiration']))
            if expiration and expiration < datetime.datetime.utcnow():
                raise AuthError('credentials have expired')

    def configure_from_aws_credential_file(self):
        if 'AWS_CREDENTIAL_FILE' in os.environ:
            path = os.getenv('AWS_CREDENTIAL_FILE')
            path = os.path.expandvars(path)
            path = os.path.expanduser(path)
            with open(path) as credfile:
                for line in credfile:
                    line = line.split('#', 1)[0]
                    if '=' in line:
                        (key, val) = line.split('=', 1)
                        if (key.strip() == 'AWSAccessKeyId' and
                                not self.args.get('key_id')):
                            # There's probably a better way to do this, but it
                            # seems to work for me.  Patches are welcome.  :)
                            self.args['key_id'] = val.strip()
                        elif (key.strip() == 'AWSSecretKey' and
                              not self.args.get('secret_key')):
                            self.args['secret_key'] = val.strip()
            return path

    def __populate_auth_args(self):
        """
        Try to get auth info from each source in turn until one provides
        both a key ID and a secret key.  After each time a source fails
        to provide enough info we wipe self.args out so we don't wind up
        mixing info from multiple sources.
        """

        # self.args gets highest precedence
        if self.args.get('key_id') and not self.args.get('secret_key'):
            # __reset_unless_ready will wipe out key_id and result in
            # the wrong error message
            raise AuthError('missing secret key; please supply one with -S')
        if self.args.get('secret_key') and not self.args.get('key_id'):
            # If only one is supplied at the command line we should
            # immediately blow up
            raise AuthError('missing access key ID; please supply one with -I')
        if self.__reset_unless_ready():
            self.log.debug('using auth info provided directly')
            return
        # Environment comes next
        self.args['key_id'] = (os.getenv('AWS_ACCESS_KEY_ID') or
                               os.getenv('AWS_ACCESS_KEY'))
        self.args['secret_key'] = (os.getenv('AWS_SECRET_ACCESS_KEY') or
                                   os.getenv('AWS_SECRET_KEY'))
        self.args['security_token'] = os.getenv('AWS_SECURITY_TOKEN')
        self.args['credential_expiration'] = \
            os.getenv('AWS_CREDENTIAL_EXPIRATION')
        if self.__reset_unless_ready():
            self.log.debug('using auth info from environment')
            return
        # See if an AWS credential file was given in the environment
        aws_credfile_path = self.configure_from_aws_credential_file()
        if aws_credfile_path and self.__reset_unless_ready():
            self.log.debug('using auth info from AWS credential file %s',
                           aws_credfile_path)
            return
        # Try the config file
        self.args['key_id'] = self.config.get_user_option('key-id')
        self.args['secret_key'] = self.config.get_user_option('secret-key',
                                                              redact=True)
        if self.__reset_unless_ready():
            self.log.debug('using auth info from configuration')
            return

    def __reset_unless_ready(self):
        """
        If both an access key ID and a secret key are set in self.args
        return True.  Otherwise, clear auth info from self.args and
        return False.
        """
        if self.args.get('key_id') and self.args.get('secret_key'):
            return True
        for arg in ('key_id', 'secret_key', 'security_token',
                    'credential_expiration'):
            self.args[arg] = None
        return False


class HmacV1Auth(HmacKeyAuth):
    '''
    S3 REST authentication
    http://docs.aws.amazon.com/AmazonS3/latest/dev/RESTAuthentication.html
    '''

    # This list comes from the CanonicalizedResource section of the above page
    HASHED_PARAMS = set((
        'acl', 'lifecycle', 'location', 'logging', 'notification',
        'partNumber', 'policy', 'requestPayment', 'torrent', 'uploadId',
        'uploads', 'versionId', 'versioning', 'versions', 'website'))

    def apply_to_request(self, req, service):
        self._update_request_before_signing(req)
        c_headers = self.get_canonicalized_headers(req)
        c_resource = self.get_canonicalized_resource(req, service)
        to_sign = self._get_string_to_sign(req, c_headers, c_resource)
        self.log.debug('string to sign: %s', repr(to_sign))
        signature = self.sign_string(to_sign.encode('utf-8'))
        self.log.debug('b64-encoded signature: %s', signature)
        self._apply_signature(req, signature)
        return req

    def apply_to_request_params(self, req, service, expiration_datetime):
        # This does not implement security tokens.
        msg = ('S3RestAuth.apply_to_request_params is deprecated; use '
               'requestbuilder.auth.aws.QueryHmacV1Auth instead')
        self.log.warn(msg)
        warnings.warn(msg, DeprecationWarning)

        for param in ('AWSAccessKeyId', 'Expires', 'Signature'):
            req.params.pop(param, None)

        expiration = calendar.timegm(expiration_datetime.utctimetuple())
        delta_t = expiration_datetime - datetime.datetime.utcnow()
        delta_t_sec = ((delta_t.microseconds +
                        (delta_t.seconds + delta_t.days * 24 * 3600) * 10**6)
                       / 10**6)
        self.log.debug('expiration: %i (%f seconds from now)',
                       expiration, delta_t_sec)
        c_headers = self.get_canonicalized_headers(req)
        self.log.debug('canonicalized headers: %s', repr(c_headers))
        c_resource = self.get_canonicalized_resource(req, service)
        self.log.debug('canonicalized resource: %s', repr(c_resource))
        to_sign = '\n'.join((req.method,
                             req.headers.get('Content-MD5', ''),
                             req.headers.get('Content-Type', ''),
                             six.text_type(expiration),
                             c_headers + c_resource))
        self.log.debug('string to sign: %s', repr(to_sign))
        signature = self.sign_string(to_sign.encode('utf-8'))
        self.log.debug('b64-encoded signature: %s', signature)
        req.params['AWSAccessKeyId'] = self.args['key_id']
        req.params['Expires'] = six.text_type(expiration)
        req.params['Signature'] = signature
        if self.args.get('security_token'):
            # This is a guess.  I have no evidence that this actually works.
            req.params['SecurityToken'] = self.args['security_token']

    def _update_request_before_signing(self, req):
        if not req.headers:
            req.headers = {}
        req.headers['Date'] = email.utils.formatdate()
        req.headers['Host'] = urlparse.urlparse(req.url).netloc
        if self.args.get('security_token'):
            req.headers['x-amz-security-token'] = self.args['security_token']
        req.headers.pop('Signature', None)

    def _get_string_to_sign(self, req, c_headers, c_resource):
        return '\n'.join((req.method.upper(),
                          req.headers.get('Content-MD5', ''),
                          req.headers.get('Content-Type', ''),
                          req.headers.get('Date'),
                          c_headers + c_resource))

    def _apply_signature(self, req, signature):
        req.headers['Authorization'] = 'AWS {0}:{1}'.format(
            self.args['key_id'], signature)

    def get_canonicalized_resource(self, req, service):
        # /bucket/keyname
        parsed_req_path = urlparse.urlparse(req.url).path
        assert service.endpoint is not None
        parsed_svc_path = urlparse.urlparse(service.endpoint).path
        # IMPORTANT:  this only supports path-style requests
        assert parsed_req_path.startswith(parsed_svc_path)
        resource = parsed_req_path[len(parsed_svc_path):]
        if parsed_svc_path.endswith('/'):
            # The leading / got stripped off
            resource = '/' + resource
        if not resource:
            # This resource does not address a bucket
            resource = '/'

        # Now append sub-resources, a.k.a. query string parameters
        if getattr(req, 'params', None):
            # A regular Request
            params = req.params
        else:
            # A PreparedRequest
            params = _get_params_from_url(req.url)
        if params:
            subresources = []
            for key, val in sorted(params.iteritems()):
                if key in self.HASHED_PARAMS:
                    if val is None:
                        subresources.append(key)
                    else:
                        subresources.append(key + '=' + val)
                if subresources:
                    resource += '?' + '&'.join(subresources)
        self.log.debug('canonicalized resource: %s', repr(resource))
        return resource

    def get_canonicalized_headers(self, req):
        headers_dict = {}
        for key, val in req.headers.iteritems():
            if key.lower().startswith('x-amz-'):
                headers_dict.setdefault(key.lower(), [])
                headers_dict[key.lower()].append(' '.join(val.split()))
        headers_strs = []
        for key, vals in sorted(headers_dict.iteritems()):
            headers_strs.append('{0}:{1}'.format(key, ','.join(vals)))
        if headers_strs:
            c_headers = '\n'.join(headers_strs) + '\n'
        else:
            c_headers = ''
        self.log.debug('canonicalized headers: %s', repr(c_headers))
        return c_headers

    def sign_string(self, to_sign):
        req_hmac = hmac.new(self.args['secret_key'], digestmod=hashlib.sha1)
        req_hmac.update(to_sign)
        return base64.b64encode(req_hmac.digest())


class QueryHmacV1Auth(HmacV1Auth):
    DEFAULT_TIMEOUT = 600  # 10 minutes

    def _update_request_before_signing(self, req):
        timeout = int(self.args.get('timeout')) or self.DEFAULT_TIMEOUT
        assert timeout > 0
        params = _get_params_from_url(req.url)
        params['AWSAccessKeyId'] = self.args['key_id']
        params['Expires'] = int(time.time() + timeout)
        params.pop('Signature', None)
        req.prepare_url(_remove_params_from_url(req.url), params)

    def _get_string_to_sign(self, req, c_headers, c_resource):
        params = _get_params_from_url(req.url)
        return '\n'.join((req.method.upper(),
                          req.headers.get('Content-MD5', ''),
                          req.headers.get('Content-Type', ''),
                          params['Expires'],
                          c_headers + c_resource))

    def _apply_signature(self, req, signature):
        req.prepare_url(req.url, {'Signature': signature})


class QueryHmacV2Auth(HmacKeyAuth):
    '''
    AWS signature version 2
    http://docs.aws.amazon.com/general/latest/gr/signature-version-2.html
    '''

    def apply_to_request(self, req, service):
        parsed = urlparse.urlparse(req.url)
        if req.method == 'POST':
            # This is probably going to break when given multipart data.
            params = urlparse.parse_qs(req.body or '', keep_blank_values=True)
        else:
            params = urlparse.parse_qs(parsed.query, keep_blank_values=True)
        params = dict((key, vals[0]) for key, vals in params.iteritems())
        params['AWSAccessKeyId'] = self.args['key_id']
        params['SignatureVersion'] = 2
        params['SignatureMethod'] = 'HmacSHA256'
        params['Timestamp'] = time.strftime(ISO8601, time.gmtime())
        if self.args.get('security_token'):
            params['SecurityToken'] = self.args['security_token']
        # Needed for retries so old signatures aren't included in to_sign
        params.pop('Signature', None)
        to_sign = '{method}\n{host}\n{path}\n'.format(
            method=req.method, host=parsed.netloc.lower(),
            path=(parsed.path or '/'))
        quoted_params = []
        for key in sorted(params):
            val = six.text_type(params[key])
            quoted_params.append(urlparse.quote(key, safe='') + '=' +
                                 urlparse.quote(val, safe='-_~'))
        query_string = '&'.join(quoted_params)
        to_sign += query_string
        # Redact passwords
        redacted_to_sign = re.sub('assword=[^&]*', 'assword=<redacted>',
                                  to_sign)
        self.log.debug('string to sign: %s', repr(redacted_to_sign))
        signature = self.sign_string(to_sign)
        self.log.debug('b64-encoded signature: %s', signature)
        params['Signature'] = signature
        if req.method == 'POST':
            req.prepare_body(params, {})
        else:
            req.prepare_url(_remove_params_from_url(req.url), params)

        return req

    def sign_string(self, to_sign):
        req_hmac = hmac.new(self.args['secret_key'], digestmod=hashlib.sha256)
        req_hmac.update(to_sign)
        return base64.b64encode(req_hmac.digest())


class HmacV4Auth(HmacKeyAuth):
    """
    AWS signature version 4
    http://docs.aws.amazon.com/general/latest/gr/signature-version-4.html
    """

    def apply_to_request(self, req, service):
        if not service.NAME:
            self.log.critical('service class %s must have a NAME attribute '
                              'to use sigv4', service.__class__.__name__)
            raise AuthError('BUG: service class {0} does not have a name'
                            .format(service.__class__.__name__))
        payload_hash = self._hash_payload(req)  # large files will be slow here
        now = time.time()
        date_header = time.strftime(ISO8601_BASIC, time.gmtime(now))
        scope = self._build_scope(service, now)
        credential = '/'.join((self.args['key_id'],) + scope)
        self._update_request_before_signing(req, credential, payload_hash,
                                            date_header)

        c_uri = self._get_canonical_uri(req)
        c_query = self._get_canonical_query(req)
        c_headers = self._get_canonical_headers(req)
        s_headers = self._get_signed_headers(req)
        c_request = '\n'.join((req.method.upper(), c_uri, c_query, c_headers,
                               '', s_headers, payload_hash))
        self.log.debug('canonical request: %s', repr(c_request))

        to_sign = '\n'.join(('AWS4-HMAC-SHA256', date_header, '/'.join(scope),
                             hashlib.sha256(c_request).hexdigest()))
        # Redact passwords
        redacted_to_sign = re.sub('assword=[^&]*', 'assword=<redacted>',
                                  to_sign)
        self.log.debug('string to sign: %s', repr(redacted_to_sign))

        derived_hmac = hmac.new('AWS4{0}'.format(self.args['secret_key']),
                                digestmod=hashlib.sha256)
        for chunk in scope:
            derived_hmac.update(chunk)
            derived_hmac = hmac.new(derived_hmac.digest(),
                                    digestmod=hashlib.sha256)
        derived_hmac.update(to_sign)
        signature = derived_hmac.hexdigest()
        self.log.debug('signature: %s', signature)
        self._apply_signature(req, credential, signature)
        return req

    def _update_request_before_signing(self, req, credential, payload_sha256,
                                       date_header):
        parsed = urlparse.urlparse(req.url)
        req.headers['Host'] = parsed.netloc
        req.headers.pop('Authorization', None)
        req.headers['X-Amz-Content-SHA256'] = payload_sha256
        req.headers['X-Amz-Date'] = date_header
        if self.args.get('security_token'):
            req.headers['X-Amz-Security-Token'] = self.args['security_token']

    def _apply_signature(self, req, credential, signature):
        auth_header = ', '.join((
            'AWS4-HMAC-SHA256 Credential={0}'.format(credential),
            'SignedHeaders={0}'.format(self._get_signed_headers(req)),
            'Signature={0}'.format(signature)))
        req.headers['Authorization'] = auth_header

    def _build_scope(self, service, timestamp):
        if service.region_name:
            region = service.region_name
        elif os.getenv('AWS_AUTH_REGION'):
            region = os.getenv('AWS_AUTH_REGION')
        else:
            self.log.error('a region name is required to use sigv4')
            raise AuthError(
                "region name is required; either use a config file "
                "to supply the service's URL or set AWS_AUTH_REGION "
                "in the environment")
        scope = (time.strftime('%Y%m%d', time.gmtime(timestamp)),
                 region, service.NAME, 'aws4_request')
        self.log.debug('scope: %s', '/'.join(scope))
        return scope

    def _get_canonical_uri(self, req):
        path = urlparse.urlsplit(req.url).path or '/'
        # TODO:  Normalize stuff like ".."
        c_uri = urlparse.quote(path, safe='/~')
        self.log.debug('canonical URI: %s', c_uri)
        return c_uri

    def _get_canonical_query(self, req):
        req_params = urlparse.parse_qsl(urlparse.urlparse(req.url).query,
                                        keep_blank_values=True)
        params = []
        for key, val in sorted(req_params or []):
            params.append('='.join((urlparse.quote(key, safe='~-_.'),
                                    urlparse.quote(val, safe='~-_.'))))
        c_params = '&'.join(params)
        self.log.debug('canonical query: %s', c_params)
        return c_params

    def _get_normalized_headers(self, req):
        # This doesn't currently support multi-value headers.
        headers = {}
        for key, val in req.headers.iteritems():
            if key.lower() not in ('connection', 'user-agent'):
                # Reverse proxies like to rewrite Connection headers.
                # Ignoring User-Agent lets us generate storable query URLs
                headers[key.lower().strip()] = val.strip()
        return headers

    def _get_canonical_headers(self, req):
        headers = []
        normalized_headers = self._get_normalized_headers(req)
        for key, val in sorted(normalized_headers.items()):
            headers.append(':'.join((key, val)))
        self.log.debug('canonical headers: %s', str(headers))
        return '\n'.join(headers)

    def _get_signed_headers(self, req):
        normalized_headers = self._get_normalized_headers(req)
        s_headers = ';'.join(sorted(normalized_headers))
        self.log.debug('signed headers: %s', s_headers)
        return s_headers

    def _hash_payload(self, req):
        if self.args.get('payload_hash'):
            return self.args['payload_hash']
        digest = hashlib.sha256()
        if not req.body:
            pass
        elif hasattr(req.body, 'seek'):
            body_position = req.data.tell()
            self.log.debug('payload hashing starting')
            while True:
                chunk = req.body.read(16384)
                if not chunk:
                    break
                digest.update(chunk)
            req.body.seek(body_position)
            self.log.debug('payload hashing done')
        elif hasattr(req.body, 'read'):
            self.log.debug('payload spooling/hashing starting')
            # 10M happens to be the size of a bundle part, the thing we upload
            # most frequently.
            spool = tempfile.SpooledTemporaryFile(max_size=(10 * 1024 * 1024))
            while True:
                chunk = req.body.read(16384)
                if not chunk:
                    break
                digest.update(chunk)
                spool.write(chunk)
            self.log.debug('payload spooling/hashing done')
            spool.seek(0)
            self.log.info('re-pointing request body at spooled payload')
            req.body = spool
            # Should we close the original req.body here?
        else:
            digest.update(req.body)
        self.log.debug('payload hash: %s', digest.hexdigest())
        return digest.hexdigest()


class QueryHmacV4Auth(HmacV4Auth):
    def _update_request_before_signing(self, req, credential, payload_sha256,
                                       date_header):
        # We don't do anything with payload_sha256.  Is that bad?
        if (req.method.upper() == 'POST' and
                'form-urlencoded' in req.headers.get('Content-Type', '')):
            self.log.warn('Query string authentication and POST form data '
                          'are generally mutually exclusive; GET is '
                          'recommended instead')
        parsed = urlparse.urlparse(req.url)
        req.headers['Host'] = parsed.netloc
        req.headers.pop('Authorization', None)
        params = {
            'X-Amz-Algorithm': 'AWS4-HMAC-SHA256',
            'X-Amz-Credential': credential,
            'X-Amz-Date': date_header,
            'X-Amz-SignedHeaders': self._get_signed_headers(req)}
        if self.args.get('timeout'):
            params['X-Amz-Expires'] = self.args['timeout']
        if self.args.get('security_token'):
            params['X-Amz-Security-Token'] = self.args['security_token']
        req.prepare_url(req.url, params)

    def _apply_signature(self, req, credential, signature):
        req.prepare_url(req.url, {'X-Amz-Signature': signature})


def _get_params_from_url(url):
    """
    Given a URL, return a dict of parameters and their values.  If a
    parameter appears more than once all but the first value will be lost.
    """
    parsed = urlparse.urlparse(url)
    params = urlparse.parse_qs(parsed.query, keep_blank_values=True)
    return dict((key, vals[0]) for key, vals in params.iteritems())


def _remove_params_from_url(url):
    """
    Return a copy of a URL with its parameters, fragments, and query
    string removed.
    """
    parsed = urlparse.urlparse(url)
    return urlparse.urlunparse((parsed[0], parsed[1], parsed[2], '', '', ''))



# Copyright (c) 2012-2013, Eucalyptus Systems, Inc.
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


class ArgumentError(ValueError):
    '''
    A bad command line argument should cause the command to fail
    '''
    pass


class ClientError(Exception):
    '''
    General client error (e.g. error accessing the server)
    '''
    pass


class AuthError(ClientError):
    '''
    Authentication handler failure
    '''
    pass


class TimeoutError(ClientError):
    """
    A request or response timeout
    """
    pass


class ServiceInitError(ClientError):
    '''
    Failure to set up a service
    '''

    def __init__(self, reason=None):
        ClientError.__init__(self, reason)


class ServerError(Exception):
    '''
    An error response from the server
    '''

    def __init__(self, response, *args):
        Exception.__init__(self, response.status_code, response.reason, *args)
        self.response = response

    @property
    def body(self):
        return self.response.text

    @property
    def reason(self):
        return self.response.reason

    @property
    def status_code(self):
        return self.response.status_code

    def format_for_cli(self):
        return 'error ({0}): {1}'.format(self.status_code,
                                         self.body or self.reason)



# Copyright (c) 2012-2015, Eucalyptus Systems, Inc.
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import argparse
import operator
import os.path
import subprocess
import warnings


__version__ = 'devel'


if '__file__' in globals():
    # Check if this is a git repo; maybe we can get more precise version info
    try:
        repo_path = os.path.join(os.path.dirname(__file__), '..')
        env = {'GIT_DIR': os.path.join(repo_path, '.git')}
        git = subprocess.Popen(['git', 'describe'], stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE, env=env)
        git.wait()
        git.stderr.read()
        if git.returncode == 0:
            __version__ = git.stdout.read().strip().lstrip('v')
            if type(__version__).__name__ == 'bytes':
                __version__ = __version__.decode()
    except:
        # Not really a bad thing; we'll just use what we had
        pass


########## SINGLETONS ##########
# Indicates a parameter that should be sent to the server without a value.
# Contrast this with empty strings, with are omitted from requests entirely.
EMPTY = type('EMPTY', (), {'__repr__': lambda self: "''",
                           '__str__': lambda self: ''})()

# Getters used for arg routing
PARAMS = operator.attrgetter('params')
SESSION = operator.attrgetter('service.session_args')


########## ARG CLASSES ##########
class Arg(object):
    '''
    A command line argument.  Positional and keyword arguments to __init__
    are the same as those to argparse.ArgumentParser.add_argument.

    The value specified by the 'dest' argument (or the one inferred if
    none is specified) is used as the name of the parameter to server
    queries unless send=False is also supplied.
    '''

    def __init__(self, *pargs, **kwargs):
        if 'route_to' in kwargs:
            if isinstance(kwargs['route_to'], (list, set, tuple)):
                self.routes = tuple(kwargs.pop('route_to'))
            else:
                self.routes = (kwargs.pop('route_to'),)
        else:
            self.routes = None
        self.pargs = pargs
        self.kwargs = kwargs

    def __eq__(self, other):
        if isinstance(other, Arg):
            return sorted(self.pargs) == sorted(other.pargs)
        return False


class MutuallyExclusiveArgList(list):
    '''
    Pass Args as positional arguments to __init__ to create a set of
    command line arguments that are mutually exclusive.  If you also
    call the required() method then the user must specify exactly one
    of them.  The recommended way to do that is via chaining it from
    __init__.

    Examples:

        MutuallyExclusiveArgList(Arg('--spam'), Arg('--eggs'))

        MutuallyExclusiveArgList(Arg('--spam'),
                                 Arg('--eggs')).required()
    '''

    def __init__(self, *args):
        if len(args) > 0 and isinstance(args[0], bool):
            # Deprecated in 0.6
            warnings.warn('positional booleans in MutuallyExclusiveArgLists '
                          'are deprecated; use .required() instead',
                          DeprecationWarning)
            self.is_required = args[0]
            list.__init__(self, args[1:])
        else:
            self.is_required = False
            list.__init__(self, args)

    def required(self):
        self.is_required = True
        return self


class Filter(object):
    '''
    An AWS query API filter.  For APIs that support filtering by
    name/value pairs, adding a Filter to a request's list of filters will
    allow a user to send an output filter to the server with '--filter
    name=value' at the command line.

    The value specified by the 'dest' argument (or the 'name' argument,
    if none is given) is used as the name of a filter in queries.
    '''
    def __init__(self, name, type=str, choices=None, help=None):
        self.name = name
        self.type = type
        self.choices = choices
        self.help = help

    def matches_argval(self, argval):
        return argval.startswith(self.name + '=')

    def convert(self, argval):
        '''
        Given an argument to --filter of the form "<name>=<value>", convert
        the value to the appropriate type by calling self.type on it, then
        return a (name, converted_value) tuple.  If the value's type conversion
        doesn't work then an ArgumentTypeError will result.  If the conversion
        succeeds but does not appear in self.choices when it exists, an
        ArgumentTypeError will result as well.
        '''
        if '=' not in argval:
            msg = "filter '{0}' must have format 'NAME=VALUE'".format(argval)
            raise argparse.ArgumentTypeError(msg)
        (name, value_str) = argval.split('=', 1)
        try:
            value = self.type(value_str)
        except ValueError:
            msg = "{0} filter value '{1}' must have type {2}".format(
                name, value_str, self.type.__name__)
            raise argparse.ArgumentTypeError(msg)
        if self.choices and value not in self.choices:
            msg = "{0} filter value '{1}' must match one of {2}".format(
                name, value,
                ', '.join([str(choice) for choice in self.choices]))
            raise argparse.ArgumentTypeError(msg)
        if value == '':
            value = EMPTY
        return (name, value)


class GenericTagFilter(Filter):
    '''
    A filter that accepts "tag:<key>=<value>" values
    '''
    def matches_argval(self, argval):
        return argval.startswith('tag:') and '=' in argval



# Copyright (c) 2012-2016 Hewlett Packard Enterprise Development LP
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import absolute_import

import itertools
import logging

import six.moves


class ConfigView(object):
    def __init__(self, data, region=None, user=None):
        self.log = data.log
        self.__data = data
        self.__region = None
        self.__user = None
        if region:
            self.region = region
        if user:
            self.user = user

    @property
    def region(self):
        return self.__region

    @region.setter
    def region(self, value):
        self.__region = value
        self.log.info('set region to %s', value)

    @property
    def user(self):
        return self.__user

    @user.setter
    def user(self, value):
        self.__user = value
        self.log.info('set user to %s', value)

    def clone(self, region=None, user=None):
        region = region or self.region
        user = user or self.user
        return ConfigView(self.__data, region=region, user=user)

    def get_global_option(self, option, redact=False):
        return self.__data.lookup_global(option, redact=redact)

    def get_region_option(self, option, region=None, redact=False):
        return self.get_region_option2(option, region=region,
                                       redact=redact)[0]

    def get_region_option2(self, option, region=None, redact=False):
        region = region or self.region
        if region:
            return self.__data.lookup(self.__data.regions, region, option,
                                      redact=redact,
                                      confdict_log_name='region')
        return None, None

    def get_user_option(self, option, user=None, redact=False):
        return self.get_user_option2(option, user=user, redact=redact)[0]

    def get_user_option2(self, option, user=None, redact=False):
        user = user or self.user
        if user:
            return self.__data.lookup(self.__data.users, user, option,
                                      redact=redact, confdict_log_name='user')
        return None, None

    def get_all_region_options(self, option):
        return self.__get_all_options(self.__data.regions, option)

    def get_all_user_options(self, option):
        return self.__get_all_options(self.__data.users, option)

    @staticmethod
    def __get_all_options(confdict, option):
        matches = {}
        for section, options in six.iteritems(confdict):
            if '*' not in section and option in options:
                matches[section] = options[option]
        return matches

    @staticmethod
    def convert_to_bool(value, default=None):
        if value is None:
            return default
        elif str(value).lower() in ('true', '1', 'yes', 'on'):
            return True
        elif str(value).lower() in ('false', '0', 'no', 'off'):
            return False
        else:
            raise ValueError('value {0} is not boolean'.format(repr(value)))


class ConfigData(object):
    def __init__(self, filenames):
        self.log = logging.getLogger('Config')
        self.log.addHandler(_NullLogHandler())  # cheap warning silencing
        self.globals = {}
        self.regions = {}
        self.users = {}
        self._memo = {}
        self._parse_config(filenames)

    def _parse_config(self, filenames):
        parser = six.moves.configparser.SafeConfigParser()
        parser.read(filenames)
        for section in parser.sections():
            if section == 'global':
                self.globals = dict(parser.items(section))
            elif section.startswith('region '):
                region = section.split()[1]
                if any(chunk == '' for chunk in region.split(':')):
                    raise ValueError(
                        'configuration file region {0} must not contain '
                        'an empty namespace'.format(repr(region)))
                if '@' in region:
                    raise ValueError(
                        'configuration file region {0} must not contain '
                        '@ characters'.format(repr(region)))
                self.regions[region] = dict(parser.items(section))
                self.regions[region].setdefault('name', region.rsplit(':')[-1])
            elif section.startswith('user '):
                user = section.split()[1]
                if any(chunk == '' for chunk in user.split(':')):
                    raise ValueError(
                        'configuration file user {0} must not contain '
                        'an empty namespace'.format(repr(user)))
                if '@' in user:
                    raise ValueError(
                        'configuration file user {0} must not contain '
                        '@ characters'.format(repr(user)))
                self.users[user] = dict(parser.items(section))
            # Ignore unrecognized sections for forward compatibility

    def lookup_global(self, option, redact=False):
        self._memo.setdefault(id(self.globals), {})
        if option in self._memo[id(self.globals)]:
            return self._memo[id(self.globals)][option]
        else:
            self.log.info('finding global option %s', option)
            value = self.globals.get(option)
            self._memo[id(self.globals)][option] = value
            if value and redact:
                self.log.info('  found   %s = <redacted>', option)
            elif value:
                self.log.info('  found   %s = %s', option, value)
            else:
                self.log.info('  novalue for %s', option)
            return value

    def lookup(self, confdict, section, option, redact=False, seen=None,
               confdict_log_name=None):
        self._memo.setdefault(id(confdict), {})
        if (section, option) in self._memo[id(confdict)]:
            return self._memo[id(confdict)][(section, option)]
        else:
            if confdict_log_name is not None:
                self.log.info('finding %s option %s', confdict_log_name,
                              option)
            values = self.__lookup(confdict, section, option, redact=redact,
                                   seen=seen)
            self._memo[id(confdict)][(section, option)] = values
            return values

    def __lookup(self, confdict, section, option, redact=False, seen=None):
        def redact_if_necessary(value):
            if redact:
                return '<redacted>'
            else:
                return value

        if seen is None:
            seen = []

        # Try wildcard combinations, from most specific to least specific,
        # except for the '*' section, which we handle later on because it is
        # also the global default.
        section_bits = section.split(':')
        for prd in itertools.product((True, False), repeat=len(section_bits)):
            if prd == (False,):
                continue  # We'll handle '*' later.
            s_section = ':'.join(section_bits[i] if prd[i] else '*'
                                 for i in range(len(section_bits)))
            self.log.debug('  section %s', s_section)
            if s_section in confdict and option in confdict[s_section]:
                value = confdict[s_section][option]
                # Check if we're redirected to another section.
                if value.startswith('use '):
                    value_bits = value.split()
                    if len(value_bits) == 1:
                        raise ValueError("something must follow 'use' in {0}"
                                         .format(value))
                    new_section = value_bits[1]
                    if len(value_bits) > 2:
                        new_option = value_bits[2]
                    else:
                        new_option = option
                    self.log.debug('  jump-to %s::%s (deferred)', new_section,
                                   new_option)
                    if new_section not in seen:
                        return self.lookup(confdict, new_section, new_option,
                                           redact=redact,
                                           seen=(seen + [section]))
                    else:
                        self.log.warning('  aborting jump due to a loop')
                # We're done!
                self.log.info('  found   %s::%s = %s', s_section, option,
                              redact_if_necessary(repr(value)))
                return value, s_section
        # Then see if we can find an exact match with a prefix in front of it.
        # Example:  'us-east-1' -> 'aws:us-east-1'
        prefixed_counts = {}
        for s_section in confdict:
            s_section_bits = s_section.split(':')
            if (len(s_section_bits) > len(section_bits) and
                    s_section_bits[-len(section_bits):] == section_bits):
                # It is longer and its last bits are the same as the entirety
                # of section_bits
                prefixed_counts.setdefault(len(s_section_bits), [])
                prefixed_counts[len(s_section_bits)].append(s_section)
        if prefixed_counts:
            shortest_prefixed = prefixed_counts[min(prefixed_counts)]
            if len(shortest_prefixed) == 1:
                self.log.debug('  jump-to %s::%s (from %s)',
                               shortest_prefixed[0], option, section)
                if shortest_prefixed[0] not in seen:
                    return self.lookup(confdict, shortest_prefixed[0], option,
                                       redact=redact, seen=(seen + [section]))
                else:
                    self.log.warning('  aborting jump due to a loop')
            else:
                raise ValueError(
                    '{0} is ambiguous; closest matches are {1}'.format(
                        repr(section), ', '.join(shortest_prefixed)))
        # Finally, try the global default, '*'.
        self.log.debug('  section *')
        if '*' in confdict and option in confdict['*']:
            self.log.info('  found   *::%s = %s', option,
                          redact_if_necessary(repr(value)))
            return value, '*'
        self.log.info('  novalue for %s', option)
        return None, None


class _NullLogHandler(logging.Handler):
    def handle(self, record):
        pass

    def emit(self, record):
        pass

    def createLock(self):
        self.lock = None



# Copyright (c) 2012-2016 Hewlett Packard Enterprise Development LP
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import absolute_import, print_function

import argparse
import bdb
import logging
import os.path
import re
import signal
import sys
import textwrap
import traceback
import warnings
try:
    import epdb
except ImportError:
    import pdb

import six

from requestbuilder import Arg, MutuallyExclusiveArgList
from requestbuilder.config import ConfigData, ConfigView
from requestbuilder.exceptions import ArgumentError
from requestbuilder.logging import configure_root_logger
from requestbuilder.suite import RequestBuilder
from requestbuilder.util import add_default_routes, aggregate_subclass_fields


class BaseCommand(object):
    """
    The basis for a command line tool.  To invoke this as a command line tool,
    call the run() method on the class.  Arguments will be parsed from the
    command line.  To invoke this in another context, such as from inside
    another command, pass keyword arguments to __init__() with names that match
    those stored by the argument parser and then call main() to retrieve a
    result.

    The general workflow of a command involves two methods:  main(), which
    inspects the arguments stored in self.args, does something, and returns a
    result; and print_result(), which takes the output of main() and prints it
    to stdout.  By default, both of these methods do nothing.  It is up to you
    to implement them to do what the tool is designed to do.

    Important members of this class include:
     - DESCRIPTION:  a string that describes the tool, which is re-wrapped and
                     used as part of the command line help string.  If it is
                     not supplied, the class's doc string, if any, will be
                     used instead.
     - USAGE:        a usage message for the command line help string.  If this
                     is None, one will be generated automatically.
     - ARGS:         a list of Arg and/or MutuallyExclusiveArgGroup objects
                     are used to generate command line arguments.  Inheriting
                     classes needing to add command line arguments should
                     contain their own ARGS lists, which are combined with
                     those of their parent classes.
    """

    DESCRIPTION = ''
    USAGE = None
    ARGS = []
    DEFAULT_ROUTES = ()
    SUITE = RequestBuilder
    __CONFIGURED_GLOBAL_STATE = False

    def __init__(self, config=None, loglevel=None, _do_cli=False, **kwargs):
        self.args = kwargs
        self.config = config  # created by _process_configfiles if None
        self.log = None  # created by _configure_logging
        self.suite = self.SUITE()
        self._arg_routes = {}  # arg name -> tuple of callables or dicts
        self._cli_parser = None  # created by _build_parser
        self.__debug = False
        self.__do_cli = _do_cli

        self._configure_logging(loglevel)
        self._process_configfiles()
        if self.__do_cli and not BaseCommand.__CONFIGURED_GLOBAL_STATE:
            self._configure_global_logging()
            BaseCommand.__CONFIGURED_GLOBAL_STATE = True

        # We need to enforce arg constraints in one location to make this
        # framework equally useful for chained commands and those driven
        # directly from the command line.  Thus, we do most of the parsing/
        # validation work before __init__ returns as opposed to putting it
        # off until we hit CLI-specific code.
        #
        # Derived classes MUST call this method to ensure things stay sane.
        self._build_parser()
        if self.__do_cli:
            # Distribute CLI args to the various places that need them
            self.process_cli_args()
        # This will raise an exception if kwargs contains anything unexpected.
        self.distribute_args()
        try:
            self.configure()
        except ArgumentError as err:
            if self.__do_cli and not self.__debug:
                # This is the only context in which we have a parser object,
                # so if we don't handle this here and now the caller will have
                # to handle it without one, which means no usage info for the
                # user.
                #
                # This is contingent on self.__debug and not self.debug because
                # run(), having no idea whether the config enables debugging or
                # not, is going to terminate the program anyway regardless of
                # that.
                self._cli_parser.error(str(err))
            else:
                raise

    @classmethod
    def from_other(cls, other, **kwargs):
        kwargs.setdefault('loglevel', other.log.level)
        new = cls(config=other.config, **kwargs)
        # That already calls configure
        return new

    def _configure_logging(self, loglevel):
        self.log = logging.getLogger(self.name)
        if loglevel is not None:
            self.log.setLevel(loglevel)
        elif self.debug:
            self.log.setLevel(logging.DEBUG)

    def _process_configfiles(self):
        if self.config is None:
            config_files = self.suite.list_config_files()
            config_data = ConfigData(config_files)
            self.config = ConfigView(config_data)
        # Now that we have a config file we should check to see if it wants
        # us to turn on debugging
        if self.debug:
            self.log.setLevel(logging.DEBUG)
            self.config.log.setLevel(logging.DEBUG)

    def _configure_global_logging(self):
        if self.config.get_global_option('debug') in ('color', 'colour'):
            configure_root_logger(use_color=True)
        else:
            configure_root_logger()

    def _build_parser(self):
        description = _rewrap_description(
            self.DESCRIPTION or self.__doc__ or '')
        parser = argparse.ArgumentParser(
            description=description, usage=self.USAGE, add_help=False,
            formatter_class=argparse.RawDescriptionHelpFormatter)
        arg_objs = self.collect_arg_objs()
        self._populate_parser(parser, arg_objs)
        # Low-level basic args that affect the core of the framework
        # These don't actually show up once CLI args finish processing.
        parser.add_argument('--debug', action='store_true', dest='_debug',
                            default=argparse.SUPPRESS,
                            help='show debugging output')
        parser.add_argument('--debugger', action='store_true',
                            dest='_debugger', default=argparse.SUPPRESS,
                            help='launch interactive debugger on error')
        parser.add_argument('--version', action='version',
                            version=self.suite.format_version(),
                            help="show the program's version and exit")
        if any('-h' in arg_obj.pargs for arg_obj in arg_objs
               if isinstance(arg_obj, Arg)):
            parser.add_argument('--help', action='help',
                                default=argparse.SUPPRESS,
                                help='show this help message and exit')
        else:
            parser.add_argument('-h', '--help', action='help',
                                default=argparse.SUPPRESS,
                                help='show this help message and exit')
        self._cli_parser = parser

    def collect_arg_objs(self):
        arg_objs = aggregate_subclass_fields(self.__class__, 'ARGS')
        add_default_routes(arg_objs, self.DEFAULT_ROUTES)
        return arg_objs

    def _populate_parser(self, parser, arg_objs):
        for arg_obj in arg_objs:
            self.__add_arg_to_cli_parser(arg_obj, parser)

    def __add_arg_to_cli_parser(self, arglike_obj, parser):
        # Returns the args the parser was populated with
        if isinstance(arglike_obj, Arg):
            if arglike_obj.kwargs.get('dest') is argparse.SUPPRESS:
                # Treat it like it doesn't exist at all
                return []
            else:
                arg = parser.add_argument(*arglike_obj.pargs,
                                          **arglike_obj.kwargs)
                if arglike_obj.routes is None:
                    self._arg_routes[arg.dest] = (None,)
                else:
                    self._arg_routes[arg.dest] = arglike_obj.routes
                return [arg]
        elif isinstance(arglike_obj, MutuallyExclusiveArgList):
            exgroup = parser.add_mutually_exclusive_group(
                required=arglike_obj.is_required)
            args = []
            for group_arg in arglike_obj:
                args.extend(self.__add_arg_to_cli_parser(group_arg, exgroup))
            return args
        elif isinstance(arglike_obj, list) or isinstance(arglike_obj, tuple):
            args = []
            for group_arg in arglike_obj:
                args.extend(self.__add_arg_to_cli_parser(group_arg, parser))
            return args
        else:
            raise TypeError('Unknown argument type ' +
                            arglike_obj.__class__.__name__)

    def process_cli_args(self):
        cli_args = vars(self._cli_parser.parse_args())
        if cli_args.pop('_debug', False):
            self.__debug = True
        if cli_args.pop('_debugger', False):
            self.__debug = True
            sys.excepthook = _debugger_except_hook
            signal.signal(signal.SIGUSR1, _debugger_usr1_handler)
        # Everything goes in self.args.  distribute_args() also puts them
        # elsewhere later on in the process.
        self.args.update(cli_args)
        redacted = type('REDACTED', (),
                        {'__repr__': lambda self: '<redacted>'})()
        for key in list(cli_args.keys()):
            if (('password' in key.lower() or 'secret' in key.lower()) and
                    cli_args[key] is not None):
                # This makes it slightly more obvious that this is redacted by
                # the framework and not just a string by removing quotes.
                cli_args[key] = redacted
        self.log.debug('parsed arguments from CLI: %s', str(cli_args))

    def distribute_args(self):
        for key, val in six.iteritems(self.args):
            # If a location to route this to was supplied, put it there, too.
            if key not in self._arg_routes:
                raise TypeError('got unrecognized arg: "{0}"'.format(key))
            routes = self._arg_routes[key]
            for route in routes:
                if route is not None:
                    if callable(route):
                        # If it's callable, call it to get the actual
                        # destination dict.  This is needed to allow Arg
                        # objects to refer to instance attributes from the
                        # context of the class.
                        route = route(self)
                    # At this point we had better have a dict.
                    route[key] = val

    def configure(self):
        # TODO:  rename this to setup
        pass

    @classmethod
    def run(cls):
        try:
            cmd = cls(_do_cli=True)
        except Exception as err:
            msg_prefix = '{0}: error:'.format(os.path.basename(sys.argv[0]))
            if isinstance(err, EnvironmentError):
                # These don't have regular 'args' attributes, and they occur
                # frequently enough they we handle them specially.
                err_bits = [msg_prefix]
                if getattr(err, 'strerror', None):
                    err_bits.append(err.strerror)
                if getattr(err, 'filename', None):
                    err_bits[-1] += ':'
                    err_bits.append(err.filename)
                print(' '.join(err_bits), file=sys.stderr)
            else:
                if len(err.args) > 0 and err.args[0]:
                    print(msg_prefix, err.args[0], file=sys.stderr)
                else:
                    print(msg_prefix, str(err), file=sys.stderr)
            # Since we don't even have a config file to consult our options for
            # determining when debugging is on are limited to what we got at
            # the command line.
            if any(arg in sys.argv for arg in ('--debug', '--debugger')):
                raise
            sys.exit(1)
        try:
            result = cmd.main()
            cmd.print_result(result)
        except Exception as err:
            cmd.handle_cli_exception(err)

    @property
    def name(self):
        return self.__class__.__name__

    def main(self):
        '''
        The main processing method.  main() is expected to do something with
        self.args and return a result.
        '''
        pass

    def print_result(self, data):
        '''
        Take a result produced by main() and print it to stdout.
        '''
        pass

    @property
    def debug(self):
        if self.__config_enables_debugging():
            return True
        if self.__debug:
            return True
        if any(arg in sys.argv for arg in ('--debug', '--debugger')):
            # In case an error occurs during argument parsing
            return True
        return False

    def handle_cli_exception(self, err):
        msg_prefix = '{0}: error:'.format(os.path.basename(sys.argv[0]))
        if isinstance(err, ArgumentError) and self.__do_cli and not self.debug:
            # Note that, unlike __init__, we get to use self.debug instead
            # of self.__debug
            self._cli_parser.error(str(err))
        if isinstance(err, EnvironmentError):
            # These don't have regular 'args' attributes, and they occur
            # frequently enough they we handle them specially.
            err_bits = [msg_prefix]
            if getattr(err, 'strerror', None):
                err_bits.append(err.strerror)
            if getattr(err, 'filename', None):
                err_bits[-1] += ':'
                err_bits.append(err.filename)
            print(' '.join(err_bits), file=sys.stderr)
        else:
            if len(err.args) > 0 and err.args[0]:
                print(msg_prefix, err.args[0], file=sys.stderr)
            else:
                print(msg_prefix, str(err), file=sys.stderr)
        if self.debug:
            raise
        sys.exit(1)

    def __config_enables_debugging(self):
        if self.config is None:
            return False
        if self.config.get_global_option('debug') in ('color', 'colour'):
            return True
        if self.config.convert_to_bool(self.config.get_global_option('debug')):
            return True
        return False


def _debugger_except_hook(type_, value, tracebk):
    '''
    Launch epdb (or pdb if epdb is unavailable) when an uncaught exception
    occurs.
    '''
    if type_ is bdb.BdbQuit:
        sys.exit(1)
    sys.excepthook = sys.__excepthook__

    if sys.stdout.isatty() and sys.stdin.isatty():
        if 'epdb' in sys.modules:
            epdb.post_mortem(tracebk, type_, value)
        else:
            pdb.post_mortem(tracebk)
    else:
        traceback.print_tb(tracebk)
        sys.exit(1)


def _debugger_usr1_handler(_, frame):
    """
    Show a traceback and local variables when sent SIGUSR1.  Note that
    this could cause exceptions due to interrupted system calls.
    """
    frame_dict = {'_frame': frame}
    frame_dict.update(frame.f_globals)
    frame_dict.update(frame.f_locals)
    print(''.join(traceback.format_stack(frame)), file=sys.stderr)


def _rewrap_description(text, width=70):
    """
    Split lines into paragraphs based on blank lines and changes in
    indentation level, rewrap the text in each of those paragraphs, and
    yield each of those paragraphs.  In paragraphs that start with
    common shell prompt characters (#, $, and %) or bullets (* and -)
    followed by whitespace, those characters do not count against the
    paragraph's indentation level.
    """
    # Match shell prompts and bullets
    bullets = {r'^(\s*)\#(\s+)': r'\1 \2',
               r'^(\s*)\$(\s+)': r'\1 \2',
               r'^(\s*)\%(\s+)': r'\1 \2',
               r'^(\s*)\*(\s+)': r'\1 \2',
               r'^(\s*)\-(\s+)': r'\1 \2'}

    def measure_margin(line):
        if not line.strip():
            return -1
        for pattern, sub in bullets.iteritems():
            line = re.sub(pattern, sub, line)
        return len(line) - len(line.lstrip())

    def find_bullet_pattern(line):
        for pattern in bullets:
            if line and re.match(pattern, line):
                return pattern

    def reformat_paragraph(lines, width=70):
        if not lines:
            return ''
        margin = measure_margin(lines[0])
        pattern = find_bullet_pattern(lines[0])
        if pattern:
            firstline_margin = len(lines[0]) - len(lines[0].lstrip())
            bullet_text = lines[0][firstline_margin:margin]
            lines[0] = re.sub(pattern, bullets[pattern], lines[0], 1)
        rewrapped = textwrap.wrap('\n'.join(line.strip() for line in lines),
                                  width=(width - margin))
        if pattern:
            combined = ([' ' * firstline_margin + bullet_text + rewrapped[0]] +
                        [' ' * margin + line for line in rewrapped[1:]])
        else:
            combined = [' ' * margin + line for line in rewrapped]

        combined = '\n'.join(' ' * margin + line for line in rewrapped)
        if pattern:
            combined = ' ' * firstline_margin + bullet_text + combined[margin:]
        return combined

    text %= dict(prog=os.path.basename(sys.argv[0]))
    in_lines = textwrap.dedent(text.rstrip().lstrip('\n')).splitlines()
    paragraphs = []
    current_lines = []
    current_margin = 0
    for line in in_lines:
        if not line.strip():
            paragraphs.append(reformat_paragraph(current_lines,
                                                 width=width))
            del current_lines[:]
            current_lines.append('')
            current_margin = -1
        else:
            pattern = find_bullet_pattern(line)
            margin = measure_margin(line)
            if pattern:
                # Bullets always begin new paragraphs
                paragraphs.append(reformat_paragraph(current_lines,
                                                     width=width))
                del current_lines[:]
            elif margin != current_margin:
                # Different margin -> start new paragraph
                paragraphs.append(reformat_paragraph(current_lines,
                                                     width=width))
                del current_lines[:]
            current_margin = margin
            current_lines.append(line)
    if current_lines:
        paragraphs.append(reformat_paragraph(current_lines, width=width))

    # Combine adjacent empty paragraphs
    for i in xrange(len(paragraphs) - 1, 0, -1):
        if paragraphs[i] == paragraphs[i - 1] == '':
            del paragraphs[i]
    return '\n'.join(paragraphs)



# Copyright (c) 2012-2016 Hewlett Packard Enterprise Development LP
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import absolute_import

import cgi
import collections
import datetime
import functools
import io
import logging
import os.path
import random
import time

import requests.exceptions
import six
import six.moves.urllib_parse as urlparse

from requestbuilder.exceptions import (ClientError, ServerError,
                                       ServiceInitError, TimeoutError)
from requestbuilder.mixins import RegionConfigurableMixin


class BaseService(RegionConfigurableMixin):
    NAME = None
    DESCRIPTION = ''
    API_VERSION = ''
    MAX_RETRIES = 2
    TIMEOUT = 30  # socket timeout in seconds

    REGION_ENVVAR = None
    URL_ENVVAR = None

    ARGS = []

    def __init__(self, config, loglevel=None, max_retries=None, timeout=None,
                 **kwargs):
        self.args = kwargs
        self.config = config
        self.endpoint = None
        self.log = logging.getLogger(self.__class__.__name__)
        if loglevel is not None:
            self.log.level = loglevel
        self.max_retries = max_retries
        self.region_name = None  # Note this can differ from config.region
        self.session_args = {}
        self.timeout = timeout
        self._session = None

    @classmethod
    def from_other(cls, other, **kwargs):
        kwargs.setdefault('loglevel', other.log.level)
        kwargs.setdefault('max_retries', other.max_retries)
        kwargs.setdefault('session_args', dict(other.session_args))
        kwargs.setdefault('timeout', other.timeout)
        if 'region' in other.args:
            kwargs.setdefault('region', other.args['region'])
        new = cls(other.config, **kwargs)
        new.configure()
        return new

    def configure(self):
        # TODO:  rename this to setup
        #
        # Configure user and region before grabbing endpoint info since
        # the latter may depend upon the former
        self.update_config_view()
        self.__configure_endpoint()

        # Configure timeout and retry handlers
        if self.max_retries is None:
            config_max_retries = self.config.get_global_option('max-retries')
            if config_max_retries is not None:
                self.max_retries = int(config_max_retries)
            else:
                self.max_retries = self.MAX_RETRIES
        if self.timeout is None:
            config_timeout = self.config.get_global_option('timeout')
            if config_timeout is not None:
                self.timeout = float(config_timeout)
            else:
                self.timeout = self.TIMEOUT

        self.session_args.setdefault('stream', True)

        # SSL cert verification is opt-in
        self.session_args.setdefault('verify', self.config.convert_to_bool(
            self.config.get_region_option('verify-ssl'), default=False))

        # Ensure everything is okay and finish up
        self.validate_config()

    @property
    def session(self):
        if self._session is None:
            self._session = requests.session()
            for key, val in six.iteritems(self.session_args):
                setattr(self._session, key, val)
            for adapter in self._session.adapters.values():
                # send_request handles retries to allow for re-signing
                adapter.max_retries = 0
        return self._session

    def validate_config(self):
        if self.endpoint is None:
            if self.NAME is not None:
                url_opt = '{0}-url'.format(self.NAME)
                available_regions = self.config.get_all_region_options(url_opt)
                if len(available_regions) > 0:
                    msg = ('No {0} endpoint to connect to was given. '
                           'Configured regions with {0} endpoints are: '
                           '{1}').format(self.NAME,
                                         ', '.join(sorted(available_regions)))
                else:
                    msg = ('No {0} endpoint to connect to was given. {0} '
                           'endpoints may be specified in a config file with '
                           '"{1}".').format(self.NAME, url_opt)
            else:
                msg = 'No endpoint to connect to was given'
            raise ServiceInitError(msg)

    def get_request_url(self, method='GET', path=None, params=None,
                        headers=None, data=None, files=None, auth=None):
        url = self.__get_url_for_path(path)

        headers = dict(headers or {})
        if 'host' not in [header.lower() for header in headers]:
            headers['Host'] = urlparse.urlparse(self.endpoint).netloc

        p_request = self.__log_and_prepare_request(method, url, params, data,
                                                   files, headers, auth)
        return p_request.url

    def send_request(self, method='GET', path=None, params=None, headers=None,
                     data=None, files=None, auth=None):
        url = self.__get_url_for_path(path)
        headers = dict(headers)
        if 'host' not in [header.lower() for header in headers]:
            headers['Host'] = urlparse.urlparse(self.endpoint).netloc

        try:
            max_tries = self.max_retries + 1
            assert max_tries >= 1
            redirects_left = 5
            if isinstance(data, file) and hasattr(data, 'seek'):
                # If we're redirected we need to be able to reset
                data_file_offset = data.tell()
            else:
                data_file_offset = None
            while True:
                for attempt_no, delay in enumerate(
                        _generate_delays(max_tries), 1):
                    # Use exponential backoff if this is a retry
                    if delay > 0:
                        self.log.debug('will retry after %.3f seconds', delay)
                        time.sleep(delay)

                    self.log.info('sending request (attempt %i of %i)',
                                  attempt_no, max_tries)
                    p_request = self.__log_and_prepare_request(
                        method, url, params, data, files, headers, auth)
                    proxies = requests.utils.get_environ_proxies(url)
                    for key, val in sorted(proxies.items()):
                        self.log.debug('request  proxy:  %s=%s', key, val)
                    p_request.start_time = datetime.datetime.now()
                    try:
                        response = self.session.send(
                            p_request, timeout=self.timeout, proxies=proxies,
                            allow_redirects=False)
                    except requests.exceptions.Timeout:
                        if attempt_no < max_tries:
                            self.log.debug('timeout', exc_info=True)
                            if data_file_offset is not None:
                                self.log.debug('re-seeking body to '
                                               'beginning of file')
                                # pylint: disable=E1101
                                data.seek(data_file_offset)
                                # pylint: enable=E1101
                                continue
                            elif not hasattr(data, 'tell'):
                                continue
                            # Fallthrough -- if it has a file pointer but not
                            # seek we can't retry because we can't rewind.
                        raise
                    if response.status_code not in (500, 503):
                        break
                    # If it *was* in that list, retry
                if (response.status_code in (301, 302, 307, 308) and
                        redirects_left > 0 and 'Location' in response.headers):
                    # Standard redirect -- we need to handle this ourselves
                    # because we have to re-sign requests when their URLs
                    # change.
                    redirects_left -= 1
                    parsed_rdr = urlparse.urlparse(
                        response.headers['Location'])
                    parsed_url = urlparse.urlparse(url)
                    new_url_bits = []
                    for rdr_bit, url_bit in zip(parsed_rdr, parsed_url):
                        new_url_bits.append(rdr_bit or url_bit)
                    if 'Host' in headers:
                        headers['Host'] = new_url_bits[1]  # netloc
                    url = urlparse.urlunparse(new_url_bits)
                    self.log.debug('redirecting to %s (%i redirect(s) '
                                   'remaining)', url, redirects_left)
                    if data_file_offset is not None:
                        self.log.debug('re-seeking body to beginning of file')
                        # pylint: disable=E1101
                        data.seek(data_file_offset)
                        # pylint: enable=E1101
                    continue
                elif response.status_code >= 300:
                    # We include 30x because we've handled the standard method
                    # of redirecting, but the server might still be trying to
                    # redirect another way for some reason.
                    self.handle_http_error(response)
                return response
        except requests.exceptions.Timeout as exc:
            self.log.debug('timeout', exc_info=True)
            raise TimeoutError('request timed out', exc)
        except requests.exceptions.ConnectionError as exc:
            self.log.debug('connection error', exc_info=True)
            return self.__handle_connection_error(exc)
        except requests.exceptions.HTTPError as exc:
            return self.handle_http_error(response)
        except requests.exceptions.RequestException as exc:
            self.log.debug('request error', exc_info=True)
            raise ClientError(exc)

    def __handle_connection_error(self, err):
        if isinstance(err, six.string_types):
            msg = err
        elif isinstance(err, Exception) and len(err.args) > 0:
            if hasattr(err.args[0], 'reason'):
                msg = err.args[0].reason
            elif isinstance(err.args[0], Exception):
                return self.__handle_connection_error(err.args[0])
            else:
                msg = err.args[0]
        else:
            raise ClientError('connection error')
        raise ClientError('connection error ({0})'.format(msg))

    def handle_http_error(self, response):
        self.log.debug('HTTP error', exc_info=True)
        raise ServerError(response)

    def __get_url_for_path(self, path):
        if path:
            # We can't simply use urljoin because a path might start with '/'
            # like it could for S3 keys that start with that character.
            if self.endpoint.endswith('/'):
                return self.endpoint + path
            else:
                return self.endpoint + '/' + path
        else:
            return self.endpoint

    def __log_and_prepare_request(self, method, url, params, data, files,
                                  headers, auth):
        hooks = {'response': functools.partial(_log_response_data, self.log)}
        if auth:
            bound_auth = auth.bind_to_service(self)
        else:
            bound_auth = None
        request = requests.Request(method=method, url=url, params=params,
                                   data=data, files=files, headers=headers,
                                   auth=bound_auth)
        p_request = self.session.prepare_request(request)
        p_request.hooks = {'response': hooks['response']}
        self.log.debug('request  method: %s', request.method)
        self.log.debug('request  url:    %s', p_request.url)
        if isinstance(p_request.headers, (dict, collections.Mapping)):
            for key, val in sorted(six.iteritems(p_request.headers)):
                if key.lower().endswith('password'):
                    val = '<redacted>'
                self.log.debug('request  header: %s: %s', key, val)
        if isinstance(request.params, (dict, collections.Mapping)):
            for key, val in sorted(urlparse.parse_qsl(
                    urlparse.urlparse(p_request.url).query,
                    keep_blank_values=True)):
                if key.lower().endswith('password'):
                    val = '<redacted>'
                self.log.debug('request  param:  %s: %s', key, val)
        if isinstance(request.data, (dict, collections.Mapping)):
            content_type, content_type_params = cgi.parse_header(
                p_request.headers.get('content-type') or '')
            if content_type == 'multipart/form-data':
                data = cgi.parse_multipart(io.BytesIO(p_request.body),
                                           content_type_params)
            elif content_type == 'application/x-www-form-urlencoded':
                data = dict(urlparse.parse_qsl(p_request.body,
                                               keep_blank_values=True))
            else:
                data = request.data
            for key, val in sorted(data.items()):
                # pylint: disable=superfluous-parens
                if key in (request.files or {}):
                    # We probably don't want to include the contents of
                    # entire files in debug output.
                    continue
                # pylint: enable=superfluous-parens
                if key.lower().endswith('password'):
                    val = '<redacted>'
                self.log.debug('request  data:   %s: %s', key, val)
        if isinstance(request.files, (dict, collections.Mapping)):
            for key, val in sorted(six.iteritems(request.files)):
                if hasattr(val, '__len__'):
                    val = '<{0} bytes>'.format(len(val))
                self.log.debug('request  file:   %s: %s', key, val)
        return p_request

    def __configure_endpoint(self):
        # self.args gets highest precedence
        if self.args.get('url'):
            url, region_name = _parse_endpoint_url(self.args['url'])
        # Environment comes next
        elif os.getenv(self.URL_ENVVAR):
            url, region_name = _parse_endpoint_url(os.getenv(self.URL_ENVVAR))
        # Try the config file
        elif self.NAME:
            url, section = self.config.get_region_option2(self.NAME + '-url')
            if section:
                # Check to see if the region name is explicitly specified
                region_name = self.config.get_region_option('name', section)
                if region_name is None:
                    # If it isn't then just grab the end of the section name
                    region_name = section.rsplit(':', 1)[-1]
            else:
                region_name = None
        self.endpoint = url
        self.region_name = region_name


def _log_response_data(logger, response, **_):
    if hasattr(response.request, 'start_time'):
        duration = datetime.datetime.now() - response.request.start_time
        logger.debug('response time:   %i.%03i seconds', duration.seconds,
                     duration.microseconds // 1000)
    if response.status_code >= 400:
        logger.error('response status: %i', response.status_code)
    else:
        logger.info('response status: %i', response.status_code)
    if isinstance(response.headers, (dict, collections.Mapping)):
        for key, val in sorted(response.headers.items()):
            logger.debug('response header: %s: %s', key, val)


def _generate_delays(max_tries):
    if max_tries >= 1:
        yield 0
        for retry_no in range(1, max_tries):
            next_delay = (random.random() + 1) * 2 ** (retry_no - 1)
            yield min((next_delay, 15))


def _parse_endpoint_url(urlish):
    """
    If given a URL, return the URL and None.  If given a URL with a string and
    "::" prepended to it, return the URL and the prepended string.  This is
    meant to give one a means to supply a region name via arguments and
    variables that normally only accept URLs.
    """
    if '::' in urlish:
        region, url = urlish.split('::', 1)
    else:
        region = None
        url = urlish
    return url, region



# Copyright (c) 2012-2016 Hewlett Packard Enterprise Development LP
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import absolute_import, print_function

import argparse
import copy
from functools import partial
import logging
import os.path
import sys
import textwrap
import warnings

import six

from requestbuilder import EMPTY, PARAMS
from requestbuilder.command import BaseCommand
from requestbuilder.exceptions import ServerError
from requestbuilder.service import BaseService
from requestbuilder.util import aggregate_subclass_fields
from requestbuilder.xmlparse import parse_listdelimited_aws_xml


class BaseRequest(BaseCommand):
    """
    The basis for a command line tool that represents a request.  The data for
    this request are stored in a few instance members:
     - method:   the HTTP method to use (e.g. 'GET').  Defaults to self.METHOD.
     - path:     the path to append to the service's path (e.g. 'sub/dir')
     - headers:  a dict of HTTP headers
     - params:   a dict of query parameters
     - body:     a string or file object containing a request body, or a dict
                 to pass to the server as form data

    This specialization of BaseCommand that implements main() as a three-step
    process:
     - preprocess():   do any processing needed before sending the request,
                       such as parsing complex command line arguments and
                       storing them in self.params, self.headers, and so forth.
     - send():         send this request to the server using the data stored
                       in its attributes, parse it using self.parse_result(),
                       and return it
     - postprocess():  given a parsed response, do any processing needed before
                       main() returns the response.

    Most requests need only implement preprocess().  Requests whose workflows
    involve sending other requests often do so in postprocess(), where the
    result of the request is known.

    Important members of this class, in addition to those inherited from
    BaseCommand, include:
     - SERVICE_CLASS:  a class corresponding to the web service in use
     - AUTH_CLASS:     a class corresponding to the authentication method
                       to use, if any
     - NAME:           a string representing the name of this request.
                       This defaults to the class's name.
     - METHOD:         the HTTP method to use by default
    """

    SERVICE_CLASS = BaseService
    AUTH_CLASS = None
    NAME = None
    METHOD = 'GET'

    DEFAULT_ROUTES = (PARAMS,)

    def __init__(self, service=None, auth=None, **kwargs):
        # Parts of the HTTP request to be sent to the server.
        self.method = self.METHOD
        self.path = None
        self.headers = {}
        self.params = {}
        self.body = ''
        self.files = {}

        # HTTP response obtained from the server
        self.response = None

        self.__configured = False
        self.__auth = auth
        self.__service = service

        BaseCommand.__init__(self, **kwargs)

    @property
    def auth(self):
        if not self.__auth and self.AUTH_CLASS is not None:
            # pylint: disable=not-callable
            self.__auth = self.AUTH_CLASS(self.config, loglevel=self.log.level)
            # pylint: enable=not-callable
        return self.__auth

    @auth.setter
    def auth(self, newval):
        self.__auth = newval

    @property
    def service(self):
        if not self.__service and self.SERVICE_CLASS is not None:
            self.__service = self.SERVICE_CLASS(self.config,
                                                loglevel=self.log.level)
        return self.__service

    @classmethod
    def from_other(cls, other, **kwargs):
        kwargs.setdefault('service', other.service)
        kwargs.setdefault('auth', other.auth)
        kwargs.setdefault('loglevel', other.log.level)
        new = cls(config=other.config, **kwargs)
        # That already calls configure
        return new

    def collect_arg_objs(self):
        arg_objs = BaseCommand.collect_arg_objs(self)
        if self.service is not None:
            arg_objs.extend(
                aggregate_subclass_fields(self.service.__class__, 'ARGS'))
        if self.auth is not None:
            arg_objs.extend(
                aggregate_subclass_fields(self.auth.__class__, 'ARGS'))
        return arg_objs

    def distribute_args(self):
        BaseCommand.distribute_args(self)
        if self.service is not None:
            self.service.args.update(self.args)
        if self.auth is not None:
            self.auth.args.update(self.args)

    def configure(self):
        # TODO:  rename this to setup
        BaseCommand.configure(self)
        if self.service is not None:
            self.service.configure()
        if self.auth is not None:
            self.auth.configure()
        self.__configured = True

    @property
    def name(self):
        return self.NAME or self.__class__.__name__

    def send(self):
        if not self.__configured:
            self.log.warn('send() called before configure(); bugs may result')
        headers = dict(self.headers or {})
        headers.setdefault('User-Agent', self.suite.get_user_agent())
        try:
            self.response = self.service.send_request(
                method=self.method, path=self.path, headers=headers,
                params=self.params, data=self.body, files=self.files,
                auth=self.auth)
            return self.parse_response(self.response)
        except ServerError as err:
            self.response = err.response
            try:
                # Empty the socket buffer so it can be reused.
                # Hopefully error responses won't be too large for this to be
                # problematic.
                if self.response is not None:
                    self.response.content
            except RuntimeError:
                # The content was already consumed
                pass
            return self.handle_server_error(err)

    def handle_server_error(self, err):
        self.log.debug('-- response content --\n', extra={'append': True})
        self.log.debug(self.response.text, extra={'append': True})
        self.log.debug('-- end of response content --')
        self.log.info('result: failure')
        raise

    def parse_response(self, response):
        return response

    def log_and_parse_response(self, response, parse_func, **kwargs):
        # We do some extra handling here to log stuff as it comes in rather
        # than reading it all into memory at once.
        self.log.debug('-- response content --\n', extra={'append': True})
        # Using Response.iter_content gives us automatic decoding, but we then
        # have to make the generator look like a file so etree can use it.
        with _IteratorFileObjAdapter(self.response.iter_content(16384)) \
                as content_fileobj:
            logged_fileobj = _ReadLoggingFileWrapper(content_fileobj, self.log,
                                                     logging.DEBUG)
            parsed_response = parse_func(logged_fileobj, **kwargs)
        self.log.debug('-- end of response content --')
        return parsed_response

    def main(self):
        '''
        The main processing method for this type of request.  In this method,
        inheriting classes generally populate self.headers, self.params, and
        self.body with information gathered from self.args or elsewhere,
        call self.send, and return the response.
        '''
        self.preprocess()
        response = self.send()
        self.postprocess(response)
        return response

    def preprocess(self):
        pass

    def postprocess(self, response):
        pass

    def handle_cli_exception(self, err):
        if isinstance(err, ServerError):
            msg = '{0}: {1}'.format(os.path.basename(sys.argv[0]),
                                    err.format_for_cli())
            print(msg, file=sys.stderr)
            if self.debug:
                raise
            sys.exit(1)
        else:
            BaseCommand.handle_cli_exception(self, err)


class AWSQueryRequest(BaseRequest):
    API_VERSION = None
    FILTERS = []
    LIST_TAGS = []

    def _populate_parser(self, parser, arg_objs):
        BaseRequest._populate_parser(self, parser, arg_objs)
        if self.FILTERS:
            parser.add_argument(
                '--filter', metavar='NAME=VALUE',
                action='append', dest='filters',
                help='restrict results to those that meet criteria',
                type=partial(_parse_filter, filter_objs=self.FILTERS))
            parser.epilog = self.__build_filter_help()
            self._arg_routes['filters'] = (None,)

    def process_cli_args(self):
        BaseRequest.process_cli_args(self)
        if 'filters' in self.args:
            self.args['Filter'] = _process_filters(self.args.pop('filters'))
            self._arg_routes['Filter'] = (self.params,)

    @property
    def action(self):
        return self.name

    def send(self):
        orig_params = self.params
        params = copy.deepcopy(self.params)
        params = self.flatten_params(self.params)
        params['Action'] = self.action
        params['Version'] = self.API_VERSION or self.service.API_VERSION
        redacted_params = dict(params)
        for key in params:
            if key.lower().endswith('password'):
                # This makes it slightly more obvious that this is redacted by
                # the framework and not just a string.
                redacted_params[key] = type(
                    'REDACTED', (), {'__repr__': lambda self: '<redacted>'})()
        self.log.info('parameters: %s', redacted_params)

        if self.method.upper() == 'POST':
            self.log.debug('sending flattened parameters as form data')
            self.body = params
            self.params = {}
        else:
            self.log.debug('sending flattened parameters as query string')
            self.params = params
        try:
            return BaseRequest.send(self)
        finally:
            self.params = orig_params

    def parse_response(self, response):
        # Parser for list-delimited responses like EC2's
        response_dict = self.log_and_parse_response(
            response, parse_listdelimited_aws_xml, list_tags=self.LIST_TAGS)
        # Strip off the root element
        assert len(response_dict) == 1
        return response_dict[list(response_dict.keys())[0]]

    def flatten_params(self, args, prefix=None):
        '''
        Given a possibly-nested dict of args and an arg routing
        destination, transform each element in the dict that matches the
        corresponding arg routing table into a simple dict containing
        key-value pairs suitable for use as query parameters.  This
        implementation flattens dicts and lists into the format given
        by AWS query APIs, which use dotted lists of dict keys and list
        indices to indicate nested structures.

        Keys with non-boolean, non-zero values that evaluate as false
        are ignored.  To include an empty string as a parameter, pass
        EMPTY (the object, not the string) as its value.

        Examples:
          in:  {'InstanceId': 'i-12345678', 'PublicIp': '1.2.3.4'}
          out: {'InstanceId': 'i-12345678', 'PublicIp': '1.2.3.4'}

          in:  {'RegionName': ['us-east-1', 'us-west-1']}
          out: {'RegionName.1': 'us-east-1',
                'RegionName.2': 'us-west-1'}

          in:  {'Filter': [{'Name':  'image-id',
                            'Value': ['ami-12345678']},
                           {'Name':  'instance-type',
                            'Value': ['m1.small', 't1.micro']}],
                'InstanceId': ['i-24680135']}
          out: {'Filter.1.Name':    'image-id',
                'Filter.1.Value.1': 'ami-12345678',
                'Filter.2.Name':    'instance-type',
                'Filter.2.Value.1': 'm1.small',
                'Filter.2.Value.2': 't1.micro',
                'InstanceId.1':     'i-24680135'}
        '''
        flattened = {}
        if args is None:
            pass
        elif isinstance(args, dict):
            for (key, val) in six.iteritems(args):
                # Prefix.Key1, Prefix.Key2, ...
                if prefix:
                    prefixed_key = '{0}.{1}'.format(prefix, key)
                else:
                    prefixed_key = str(key)

                if isinstance(val, dict) or isinstance(val, list):
                    flattened.update(self.flatten_params(val, prefixed_key))
                elif isinstance(val, bool):
                    flattened[prefixed_key] = str(val).lower()
                elif isinstance(val, file):
                    flattened[prefixed_key] = val.read()
                elif isinstance(val, float):
                    flattened[prefixed_key] = str(val)
                elif val or val is 0:
                    flattened[prefixed_key] = str(val)
                elif val is EMPTY:
                    flattened[prefixed_key] = ''
        elif isinstance(args, list):
            for (i_item, item) in enumerate(args, 1):
                # Prefix.1, Prefix.2, ...
                if prefix:
                    prefixed_key = '{0}.{1}'.format(prefix, i_item)
                else:
                    prefixed_key = str(i_item)

                if isinstance(item, dict) or isinstance(item, list):
                    flattened.update(self.flatten_params(item, prefixed_key))
                elif isinstance(item, bool):
                    flattened[prefixed_key] = str(item).lower()
                elif isinstance(item, file):
                    flattened[prefixed_key] = item.read()
                elif item or item == 0:
                    flattened[prefixed_key] = str(item)
                elif item is EMPTY:
                    flattened[prefixed_key] = ''
        else:
            raise TypeError('non-flattenable type: ' + args.__class__.__name__)
        return flattened

    def __build_filter_help(self, force=False):
        '''
        Return a pre-formatted help string for all of the filters defined in
        self.FILTERS.  The result is meant to be used as command line help
        output.
        '''
        # Does not have access to self.config
        if '-h' not in sys.argv and '--help' not in sys.argv and not force:
            # Performance optimization
            return ''

        helplines = ['allowed filter names:']
        for filter_obj in self.FILTERS:
            if filter_obj.help == argparse.SUPPRESS:
                continue
            elif filter_obj.help:
                first, _, rest = filter_obj.help.partition('\n')
                if rest.startswith(' ') and not first.startswith(' '):
                    # First line is not uniformly indented
                    content = first + ' ' + textwrap.dedent(rest)
                else:
                    content = filter_obj.help
                if len(filter_obj.name) <= 20:
                    # Short name; start on same line and pad two spaces
                    firstline = '  {0:<20}  '.format(filter_obj.name)
                    wrapper = textwrap.TextWrapper(
                        fix_sentence_endings=True, initial_indent=firstline,
                        subsequent_indent=(' ' * 24))
                else:
                    # Long name; start on next line
                    helplines.append('  ' + filter_obj.name)
                    wrapper = textwrap.TextWrapper(
                        fix_sentence_endings=True, initial_indent=(' ' * 24),
                        subsequent_indent=(' ' * 24))
                helplines.extend(wrapper.wrap(content))
            else:
                # No help; everything goes on one line
                helplines.append('  ' + filter_obj.name)
        return '\n'.join(helplines)


def _parse_filter(filter_str, filter_objs=None):
    '''
    Given a "key=value" string given as a command line parameter, return a pair
    with the matching filter's dest member and the given value after converting
    it to the type expected by the filter.  If this is impossible, an
    ArgumentTypeError will result instead.
    '''
    # Find the appropriate filter object
    filter_objs = [obj for obj in (filter_objs or [])
                   if obj.matches_argval(filter_str)]
    if not filter_objs:
        msg = '"{0}" matches no available filters'.format(filter_str)
        raise argparse.ArgumentTypeError(msg)
    return filter_objs[0].convert(filter_str)


def _process_filters(cli_filters):
    '''
    Change filters from the [(key, value), ...] format given at the command
    line to [{'Name': key, 'Value': [value, ...]}, ...] format, which
    flattens to the form the server expects.
    '''
    filter_args = {}
    # Compile [(key, value), ...] pairs into {key: [value, ...], ...}
    for (key, val) in cli_filters or {}:
        filter_args.setdefault(key, [])
        filter_args[key].append(val)
    # Build the flattenable [{'Name': key, 'Value': [value, ...]}, ...]
    filters = [{'Name': name, 'Value': values} for (name, values)
               in six.iteritems(filter_args)]
    return filters


class _IteratorFileObjAdapter(object):
    def __init__(self, source):
        self._source = source
        self._buflist = []
        self._closed = False
        self._len = 0

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    @property
    def closed(self):
        return self._closed

    def close(self):
        if not self._closed:
            self._buflist = None
            self._closed = True

    def read(self, size=-1):
        if size is None or size < 0:
            for chunk in self._source:
                self._buflist.append(chunk)
            result = ''.join(self._buflist)
            self._buflist = []
            self._len = 0
        else:
            while self._len < size:
                try:
                    chunk = next(self._source)
                    self._buflist.append(chunk)
                    self._len += len(chunk)
                except StopIteration:
                    break
            result = ''.join(self._buflist)
            extra_len = len(result) - size
            self._buflist = []
            self._len = 0
            if extra_len > 0:
                self._buflist = [result[-extra_len:]]
                self._len = extra_len
                result = result[:-extra_len]
        return result


class _ReadLoggingFileWrapper(object):
    def __init__(self, fileobj, logger, level):
        self.fileobj = fileobj
        self.logger = logger
        self.level = level

    def read(self, size=-1):
        chunk = self.fileobj.read(size)
        self.logger.log(self.level, chunk, extra={'append': True})
        return chunk



# Copyright (c) 2012-2014, Eucalyptus Systems, Inc.
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from distutils.command.build_py import build_py
from distutils.command.sdist import sdist
import os.path
import sys

from setuptools import find_packages, setup

from requestbuilder import __version__


requirements = ['requests>=2.4',
                'six']
if sys.version_info < (2, 7):
    requirements.append('argparse')


class build_py_with_git_version(build_py):
    '''Like build_py, but also hardcoding the version in __init__.__version__
       so it's consistent even outside of the source tree'''

    def build_module(self, module, module_file, package):
        build_py.build_module(self, module, module_file, package)
        if module == '__init__' and '.' not in package:
            version_line = "__version__ = '{0}'\n".format(__version__)
            old_init_name = self.get_module_outfile(self.build_lib, (package,),
                                                    module)
            new_init_name = old_init_name + '.new'
            with open(new_init_name, 'w') as new_init:
                with open(old_init_name) as old_init:
                    for line in old_init:
                        if line.startswith('__version__ ='):
                            new_init.write(version_line)
                        else:
                            new_init.write(line)
                new_init.flush()
            os.rename(new_init_name, old_init_name)


class sdist_with_git_version(sdist):
    '''Like sdist, but also hardcoding the version in __init__.__version__ so
       it's consistent even outside of the source tree'''

    def make_release_tree(self, base_dir, files):
        sdist.make_release_tree(self, base_dir, files)
        version_line = "__version__ = '{0}'\n".format(__version__)
        old_init_name = os.path.join(base_dir, 'requestbuilder/__init__.py')
        new_init_name = old_init_name + '.new'
        with open(new_init_name, 'w') as new_init:
            with open(old_init_name) as old_init:
                for line in old_init:
                    if line.startswith('__version__ ='):
                        new_init.write(version_line)
                    else:
                        new_init.write(line)
            new_init.flush()
        os.rename(new_init_name, old_init_name)


setup(name='requestbuilder',
      version=__version__,
      description='Command line-driven HTTP request builder',
      author='Garrett Holmstrom (gholms)',
      author_email='gholms@devzero.com',
      url='https://github.com/boto/requestbuilder',
      packages=find_packages(),
      install_requires=requirements,
      license='ISC',
      platforms='Posix; MacOS X',
      classifiers=['Development Status :: 4 - Beta',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: ISC License (ISCL)',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python',
                   'Programming Language :: Python :: 2',
                   'Programming Language :: Python :: 2.6',
                   'Programming Language :: Python :: 2.7',
                   'Topic :: Internet'],
      cmdclass={'build_py': build_py_with_git_version,
                'sdist': sdist_with_git_version})



# Copyright (c) 2012-2015, Eucalyptus Systems, Inc.
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import argparse
import math
import signal
import sys
import time

try:
    import progressbar
except ImportError:
    pass

from requestbuilder import Arg, MutuallyExclusiveArgList


if 'progressbar' in sys.modules:
    _PROGRESS_BAR_COMMAND_ARGS = [
        MutuallyExclusiveArgList(
            Arg('--progress', dest='show_progress', action='store_true',
                default=sys.stdout.isatty(), route_to=None,
                help='show progress (the default when run interactively)'),
            Arg('--no-progress', dest='show_progress', action='store_false',
                default=sys.stdout.isatty(), route_to=None, help='''do not
                show progress (the default when run non-interactively)'''),
            Arg('--porcelain', dest='show_porcelain', action='store_true',
                route_to=None, help=argparse.SUPPRESS))]
else:
    # Keep them around so scripts don't break, but make them non-functional
    #
    # This isn't in a MutuallyExclusiveArgList because of an argparse bug:
    # http://bugs.python.org/issue17890
    _PROGRESS_BAR_COMMAND_ARGS = [
        Arg('--progress', dest='show_progress', action='store_false',
            default=False, route_to=None, help=argparse.SUPPRESS),
        Arg('--no-progress', dest='show_progress', action='store_false',
            default=False, route_to=None, help=argparse.SUPPRESS),
        Arg('--porcelain', dest='show_porcelain', action='store_true',
            route_to=None, help=argparse.SUPPRESS)]


class FileTransferProgressBarMixin(object):
    '''
    A command mixin that provides download/upload progress bar support,
    along with options to enable or disable them.  If progress bars are
    disabled at the command line get_progressbar will return None.  If the
    progressbar module is unavailable get_progressbar will return None *and*
    no progress-related options will be added.
    '''

    ARGS = _PROGRESS_BAR_COMMAND_ARGS

    def get_progressbar(self, label=None, maxval=None):
        if self.args.get('show_porcelain'):
            return _MachineReadableCounter(label=label, maxval=maxval)
        elif 'progressbar' in sys.modules and self.args.get('show_progress',
                                                            False):
            widgets = []
            if label is not None:
                widgets += [label, ' ']
            if maxval is not None:
                widgets += [progressbar.Percentage(), ' ',
                            progressbar.Bar(marker='='), ' ',
                            _FileSize(), ' ',
                            progressbar.FileTransferSpeed(), ' ']
                if 'AdaptiveETA' in dir(progressbar):
                    widgets.append(progressbar.AdaptiveETA())
                else:
                    widgets.append(progressbar.ETA())
                pbar = progressbar.ProgressBar(widgets=widgets,
                                               maxval=(maxval or sys.maxint),
                                               poll=0.05)
                #
                # The ProgressBar class initializer installs a signal handler
                # for SIGWINCH to resize the progress bar. Sometimes this can
                # interrupt long running system calls which can cause an
                # IOError exception to be raised. The call to siginterrupt
                # below will retrieve the currently installed signal handler
                # for SIGWINCH and set the SA_RESTART flag. This will cause
                # system calls to be restarted after the handler has been
                # executed instead of raising an exception.
                #
                signal.siginterrupt(signal.SIGWINCH, False)
                return pbar
            else:
                widgets += [_IndeterminateBouncingBar(marker='='), ' ',
                            _FileSize(), ' ',
                            progressbar.FileTransferSpeed(), ' ',
                            progressbar.Timer(format='Time: %s')]
                pbar = _IndeterminateProgressBar(widgets=widgets,
                                                 maxval=(maxval or sys.maxint),
                                                 poll=0.05)
                # See comment above
                signal.siginterrupt(signal.SIGWINCH, False)
                return pbar
        else:
            return _EveryMethodObject()


# Used as a placeholder for ProgressBar when progressbar isn't there
class _EveryMethodObject(object):
    def do_nothing(self, *args, **kwargs):
        pass

    def __getattribute__(self, name):
        return object.__getattribute__(self, 'do_nothing')


if 'progressbar' in sys.modules:
    class _IndeterminateProgressBar(progressbar.ProgressBar):
        def finish(self):
            self.maxval = self.currval
            progressbar.ProgressBar.finish(self)

    class _IndeterminateBouncingBar(progressbar.BouncingBar):
        '''
        A BouncingBar that moves exactly one space each time it updates,
        rather than one space per unit.  This is mainly used for downloads with
        unknown lengths.
        '''
        def __init__(self, *args, **kwargs):
            progressbar.BouncingBar.__init__(self, *args, **kwargs)
            self.__update_count = 0

        def update(self, pbar, width):
            orig_currval = pbar.currval
            pbar.currval = self.__update_count
            retval = progressbar.BouncingBar.update(self, pbar, width)
            pbar.currval = orig_currval
            self.__update_count += 1
            return retval

    class _FileSize(progressbar.Widget):
        PREFIXES = ' kMGTPEZY'

        def update(self, pbar):
            if pbar.currval == 0:
                power = 0
                scaledval = 0
            else:
                power = int(math.log(pbar.currval, 1024))
                scaledval = pbar.currval / 1024.0 ** power
            return '{0:6.2f} {1}B'.format(scaledval, self.PREFIXES[power])


class _MachineReadableCounter(object):
    def __init__(self, maxval=None, label=None):
        self.maxval = maxval
        self.currval = 0
        self._last_displayed_val = None
        self._last_updated = 0
        self._finished = False
        if label:
            self.__template = '{0} '.format(label)
        else:
            self.__template = ''
        if self.maxval:
            self.__template = '{0}{{0}}/{1}\n'.format(self.__template,
                                                      int(self.maxval))
        else:
            self.__template = '{0}{{0}}\n'.format(self.__template)

    def start(self):
        self._display()

    def update(self, val):
        self.currval = val
        delta = time.time() - self._last_updated
        if (delta > 0.1 and self.currval != self._last_displayed_val and
                not self._finished):
            self._display()
            self._last_updated = time.time()

    def finish(self):
        if self.maxval:
            self.currval = self.maxval
        self._display()
        self._finished = True

    def _display(self):
        sys.stderr.write(self.__template.format(int(self.currval)))
        self._last_displayed_val = self.currval



# Copyright (c) 2012-2015, Eucalyptus Systems, Inc.
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import absolute_import

import logging


NOTICE = (logging.WARN + logging.INFO) // 2


class ProgressiveStreamHandler(logging.StreamHandler):
    '''
    A handler class that allows the "cursor" to stay on one line for selected
    messages
    '''

    def __init__(self, **kwargs):
        logging.StreamHandler.__init__(self, **kwargs)
        self.appending = False

    def emit(self, record):
        try:
            if getattr(record, 'append', False):
                if self.appending:
                    self.stream.write(record.getMessage())
                else:
                    self.stream.write(self.format(record))
                self.appending = True
            else:
                terminator = getattr(self, 'terminator', '\n')
                if self.appending:
                    self.stream.write(terminator)
                self.stream.write(self.format(record))
                self.stream.write(terminator)
                self.appending = False
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            self.handleError(record)


class ColoringFormatter(logging.Formatter):
    LOG_COLORS = [(logging.ERROR, '\033[91m'),
                  (logging.WARN, '\033[93m'),
                  (NOTICE, '\033[92m'),
                  (logging.INFO, '\033[96m'),
                  (logging.DEBUG, '\033[94m')]

    def format(self, record):
        msg = logging.Formatter.format(self, record)
        for level, colorcode in sorted(self.LOG_COLORS, reverse=True):
            if record.levelno >= level:
                return colorcode + msg + '\033[0m'
        return msg


class RequestbuilderLogger(logging.Logger):
    def notice(self, msg, *args, **kwargs):
        self.log(NOTICE, msg, *args, **kwargs)


def configure_root_logger(use_color=False):
    logging.addLevelName(NOTICE, 'NOTICE')
    logging.setLoggerClass(RequestbuilderLogger)
    logfmt = '%(asctime)s %(levelname)-7s %(name)s %(message)s'
    rootlogger = logging.getLogger('')
    handler = ProgressiveStreamHandler()
    if use_color:
        formatter = ColoringFormatter(logfmt)
    else:
        formatter = logging.Formatter(logfmt)
    handler.setFormatter(formatter)
    rootlogger.addHandler(handler)
    rootlogger.setLevel(100)

    # Attempt to have logging capture warnings as well (requires 2.7)
    try:
        logging.captureWarnings(True)
    except AttributeError:
        pass
